package de.tud.kom.p2psim.impl.topology.movement.vehicular.sumo.simulation.controller.traci;

import de.tud.kom.p2psim.api.simengine.SimulatorObserver;
import de.tud.kom.p2psim.impl.simengine.Simulator;
import de.tud.kom.p2psim.impl.topology.PositionVector;
import de.tud.kom.p2psim.impl.topology.movement.vehicular.sumo.simulation.controller.VehicleInformationContainer;
import de.tudarmstadt.maki.simonstrator.api.Randoms;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.api.EdgeController;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.api.SimulationSetupExtractor;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.api.TrafficLightController;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.api.VehicleController;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.api.information.LocationUtils;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.*;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.exception.NoAdditionalRouteAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.exception.NoExitAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.routing.DijkstraAlgorithm;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.routing.RoutingAlgorithm;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.trafficlights.TrafficLight;
import de.tudarmstadt.maki.simonstrator.api.component.vehicular.roadnetwork.trafficlights.TrafficLightState;
import de.tudresden.sumo.cmd.*;
import de.tudresden.sumo.util.SumoCommand;
import de.tudresden.ws.container.*;
import it.polito.appeal.traci.SumoTraciConnection;

import java.io.*;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.*;

/**
 *
 * @author Tobias Meuser (tobias.meuser@kom.tu-darmstadt.de)
 * @version 1.0 at 06.11.2017
 *
 */
public class TraciSimulationController implements VehicleController, SimulationSetupExtractor, EdgeController, SimulatorObserver, TrafficLightController {
	private static final File TEMP_FILE = new File(new File(System.getProperty("java.io.tmpdir")), "road_network.tmp");

	private Random _random = Randoms.getRandom(getClass());

	private static final Map<String, TraciSimulationController> CONTROLLER = new HashMap<>();

	private static final double CLUSTERING_DISTANCE = 50;

	private String _sumoExe;
	private String _configFile;
	private String _tripOutputFile;
	private int _seed;

	private SumoTraciConnection _connection;

	private double _start = -1;
	private double _step = -1;

	private double _startX;
	private double _startY;
	private double _endX;
	private double _endY;

	private Map<String, VehicleInformationContainer> _positons = new HashMap<>();

	private boolean _initalized = false;

	private boolean _observedAreaSet;

	private Map<String, Double> _vehiclesOutOfRange = new HashMap<>();

	private RoadNetwork _roadNetwork;

	private RoutingAlgorithm _algorithm = new DijkstraAlgorithm();

	public static synchronized TraciSimulationController createSimulationController(String pSumoExe, String pConfigFile, String pTripOutputFile) {
		if (!CONTROLLER.containsKey(pConfigFile)) {
			CONTROLLER.put(pConfigFile, new TraciSimulationController(pSumoExe, pConfigFile, pTripOutputFile));
		}
		return CONTROLLER.get(pConfigFile);
	}

	public static synchronized TraciSimulationController createSimulationController(String pSumoExe, String pConfigFile, String pTripOutputFile, int seed) {
		if (!CONTROLLER.containsKey(pConfigFile)) {
			CONTROLLER.put(pConfigFile, new TraciSimulationController(pSumoExe, pConfigFile, pTripOutputFile, seed));
		}
		return CONTROLLER.get(pConfigFile);
	}

	private TraciSimulationController(String pSumoExe, String pConfigFile, String pTripOutputFile) {
		_sumoExe = pSumoExe;
		_configFile = pConfigFile;
		_tripOutputFile = pTripOutputFile;
	}

	private TraciSimulationController(String pSumoExe, String pConfigFile, String pTripOutputFile, int seed) {
		_sumoExe = pSumoExe;
		_configFile = pConfigFile;
		_tripOutputFile = pTripOutputFile;
		_seed = seed;
	}


	public static VehicleController getSimulationController() {
		return CONTROLLER.values().iterator().next();
	}

	public static TrafficLightController getTrafficLightController() {
		return CONTROLLER.values().iterator().next();
	}

	@Override
	public synchronized void init() {
		if (!_initalized) {
			Random random = Randoms.getRandom("SUMO");

			// This will only work with the updated version of the TraaS API for sumo
			// It is available for download at https://dev.kom.e-technik.tu-darmstadt.de/gitlab/tobiasm/TraaS.git
			int nextRandomInt = random.nextInt();
			if (_seed > 0) {
				nextRandomInt = _seed;
			}
			_connection = new SumoTraciConnection(_sumoExe, _configFile, nextRandomInt);
			/*
			 * prevent vehicles form teleporting (http://sumo.dlr.de/wiki/Simulation/Why_Vehicles_are_teleporting)
			 */

			if (_tripOutputFile.contains("basic")) {
			    _connection.addOption("additional-files", getAdditionalsFilename("basic-additional.xml"));
            }

			_connection.addOption("time-to-teleport", Integer.toString(-1));
			_connection.addOption("tripinfo-output.write-unfinished", "t");
			String outputTripFileName = getTripOutputFilename(_tripOutputFile);
			String summaryFile = getSummaryFilename(_tripOutputFile);
			_connection.addOption("tripinfo-output", outputTripFileName);
			_connection.addOption("device.emissions.probability", "1.0");
			_connection.addOption("duration-log.statistics", "t");
			_connection.addOption("summary", summaryFile);
			_connection.printSumoError(true);

			try {
				_connection.runServer();
			} catch (RuntimeException e) {
				throw e;
			} catch (Exception e) {
				e.printStackTrace();
			}

			Simulator.getInstance().addObserver(this);

			_initalized = true;
		}
	}

	private String getSummaryFilename(String tripOutputFile) {
		Path path = Paths.get(_configFile);
		return  path.getParent().toString() + "/summary/" + tripOutputFile;
	}

	private String getTripOutputFilename(String filename) {
		Path path = Paths.get(_configFile);
		return path.getParent().toString() + "/results/" + filename;
	}

	private String getAdditionalsFilename(String filename) {
		Path path = Paths.get(_configFile);
		return path.getParent().toString() + "/" + filename;
	}

	@Override
	public void simulationFinished() {
		/*
		 * This is called by the simulation scheduler once the simulation is
		 * finished - it should be used to terminate the connection to SUMO.
		 *
		 */
		if (_connection != null && !_connection.isClosed()) {
			_connection.close();
		}
	}

	@Override
	public double getVehicleWaitingTime(String vehicleId) {
		SumoCommand waitingTimeCommand = Vehicle.getWaitingTime(vehicleId);
		Object waitingTimeObject = requestObject(waitingTimeCommand);
		double waitingTime = (double) waitingTimeObject;

		return waitingTime;
	}

	@Override
	public Location getVehiclePosition(String pVehicleID) {
		return _positons.get(pVehicleID).getPosition();
	}

    @Override
    public double getVehicleMaxSpeed(String vehicleId) {
        SumoCommand speedCommand = Vehicle.getMaxSpeed(vehicleId);
        Object object = requestObject(speedCommand);

        return (Double)object;
    }

	@Override
	public int getEdgeOccupancyRate(String edgeIdFrom) {
		SumoCommand edgeCommand = Edge.getLastStepOccupancy(edgeIdFrom);
		Object edgeObject = requestObject(edgeCommand);

		int edgeOccupancyRate = (int)Math.round(((Double)edgeObject)*100.0);

		return edgeOccupancyRate;

	}

    @Override
	public double getVehicleHeading(String pVehicleID) {
		return _positons.get(pVehicleID).getHeading();
	}

	@Override
	public String getVehicleCurrentStreet(String vehicleId) {
		SumoCommand currentRoadCommand = Vehicle.getRoadID(vehicleId);
		Object currentRoadObject = requestObject(currentRoadCommand);
		String currentRoad = (String) currentRoadObject;

		return currentRoad;
	}

	@Override
	public double getVehicleLength(String vehicleID) {
		SumoCommand vehicleLengthCommand = Vehicle.getLength(vehicleID);
		Object vehicleLengthObject = requestObject(vehicleLengthCommand);
		double length = (double) vehicleLengthObject;

		return length;
	}

	@Override
	public Location getVehiclePosition(double pStep, String pVehicleID) {
		if (pStep == _step) {
			return getVehiclePosition(pVehicleID);
		}
		throw new UnsupportedOperationException("Future locations is not supported anymore!");
	}

	@Override
	public List<Location> getAllIntersections(boolean pCluster) {
		List<Location> result = new ArrayList<>();

		SumoCommand intersectionCommand = Junction.getIDList();

		Object intersectionObject = requestObject(intersectionCommand);

		SumoStringList intersections = (SumoStringList) intersectionObject;

		for (String intersection : intersections) {
			SumoCommand positionCommand = Junction.getPosition(intersection);
			Object positionObject = requestObject(positionCommand);
			SumoPosition2D sumoPosition = (SumoPosition2D) positionObject;

			if (_observedAreaSet) {
				if (_startX <= sumoPosition.x && sumoPosition.x <= _endX && _startY <= sumoPosition.y && sumoPosition.y <= _endY) {
					result.add(new PositionVector(sumoPosition.x - _startX, sumoPosition.y - _startY, 0));
				}
			} else {
				result.add(new PositionVector(sumoPosition.x, sumoPosition.y, 0));
			}
		}

		if (pCluster) {
			List<Location> tempResult = new ArrayList<>();

			outer:for (int i = 0; i < result.size(); i++) {
				Location position = result.get(i);

				for (int j = 0; j < tempResult.size(); j++) {
					Location addedPosition = tempResult.get(j);
					if (position != addedPosition && LocationUtils.calculateDistance(position, addedPosition) < CLUSTERING_DISTANCE / 1000.0) {
						continue outer;
					}
				}

				tempResult.add(position);
			}

			result = tempResult;
		}

		return result;
	}

	@Override
	public int getPhase(String trafficLightId) {
		SumoCommand phaseCommand = Trafficlight.getPhase(trafficLightId);
		Object phaseObject = requestObject(phaseCommand);
		int phase = (int)phaseObject;

		return phase;
	}

	@Override
	public void setPhase(String trafficLightId, int phase) {
		SumoCommand phaseCommand = Trafficlight.setPhase(trafficLightId, phase);
		execute(phaseCommand);
	}

	@Override
	public long getRemainingPhaseTime(String trafficLightId) {
		SumoCommand nextSwitchCommand = Trafficlight.getNextSwitch(trafficLightId);
		Object nextSwitchObject = requestObject(nextSwitchCommand);
		Object currentTime = requestObject(Simulation.getCurrentTime());

		long remaingPhaseTime = (long)(((Double) nextSwitchObject).longValue() - (((Integer)currentTime).longValue()/1000.0));

		return remaingPhaseTime;
	}

    @Override
    public void setPhaseDuration(String trafficLightId, int extraTime) {
        SumoCommand phaseDurationCommand = Trafficlight.setPhaseDuration(trafficLightId, extraTime);
        execute(phaseDurationCommand);
    }

	@Override
	public void setPhaseDuration(String trafficLightId, double extraTime) {
		SumoCommand phaseDurationCommand = Trafficlight.setPhaseDuration(trafficLightId, extraTime);
		execute(phaseDurationCommand);
	}

	@Override
	public double getPhaseDuration(String trafficLightId) {
		SumoCommand phaseDurationCommand  = Trafficlight.getPhaseDuration(trafficLightId);
		Object phaseDurationObject = requestObject(phaseDurationCommand);
		double phaseDuration = ((double) phaseDurationObject)/1000.0;

		return phaseDuration;
	}

	@Override
	public int getProgramId(String trafficLightId) {
		SumoCommand programIdCommand = Trafficlight.getProgram(trafficLightId);
		Object programIdObject = requestObject(programIdCommand);
		int programId = Integer.parseInt((String)programIdObject);

		return programId;
	}

    @Override
    public void setCompleteRedYellowGreenDefinition(String trafficLightId, SumoTLSProgram sumoTLSProgram) {
	    SumoCommand definitionCommand = Trafficlight.setCompleteRedYellowGreenDefinition(trafficLightId, sumoTLSProgram);
	    execute(definitionCommand);
    }

	@Override
	public boolean isValidTrafficLightId(String nextTL) {
		boolean isValid = false;

		SumoCommand trafficLightCommand = Trafficlight.getIDList();
		Object object = requestObject(trafficLightCommand);
		SumoStringList trafficLightList = (SumoStringList)object;

		for (String trafficLightId : trafficLightList) {
			if (trafficLightId.equalsIgnoreCase(nextTL)) {
				isValid = true;
				break;
			}
		}

		return isValid;
	}

	@Override
	public double getRoadLength(String trafficLightId, String roadId) {
		double length = 0.0;
		int amountOfLanes = 0;
		List<String> lanes = getControlledLanes(trafficLightId);
		for (String laneId : lanes) {
			if (laneId.contains(roadId)) {
				SumoCommand speedCommand = Lane.getLength(laneId);
				Object object = requestObject(speedCommand);
				length += (double) object;
				amountOfLanes++;
			}
		}

		return length / (double) amountOfLanes;
	}

	@Override
	public double getVehicleAcceleration(String vehicleId) {
	    SumoCommand accelerationCommand = Vehicle.getAccel(vehicleId);
	    Object accelerationObject = requestObject(accelerationCommand);
	    double vehicleAcceleration = (double) accelerationObject;

	    return vehicleAcceleration;
	}

	@Override
	public String getDefinition(String trafficLightId) {
		SumoCommand definitionCommand = Trafficlight.getCompleteRedYellowGreenDefinition(trafficLightId);
		Object definitionObject = requestObject(definitionCommand);

		//((SumoTLSController) definitionObject).programs.get("1")

		return (String) definitionObject;
	}

	@Override
	public int getLastStepVehicleNumber(String laneIdFrom) {
		SumoCommand lastStepVehicleNumberCommand = Lane.getLastStepVehicleNumber(laneIdFrom);
		Object lastStepVehicleNumberObject = requestObject(lastStepVehicleNumberCommand);
		int lastStepVehicleNumber = ((Integer)lastStepVehicleNumberObject).intValue();

		return lastStepVehicleNumber;
	}

	@Override
	public SumoTLSController getCompleteDefinition(String trafficLightId) {
		SumoCommand completeTrafficLightDefinitionCommand = Trafficlight.getCompleteRedYellowGreenDefinition(trafficLightId);
		Object completeTrafficLightDefinitionObject = requestObject(completeTrafficLightDefinitionCommand);

		return  ((SumoTLSController) completeTrafficLightDefinitionObject);
	}

	@Override
	public List<String> getLastStepVehicleIDs(String laneId) {
		SumoCommand lastStepVehicleIDsCommand = Lane.getLastStepVehicleIDs(laneId);
		Object lastStepVehicleIDsObject = requestObject(lastStepVehicleIDsCommand);
		List<String> lastStepVehicleIDs = (List<String>) lastStepVehicleIDsObject;

		return lastStepVehicleIDs;
	}

	@Override
	public boolean nextStep() {
		if (Simulator.getEndTime() == Simulator.getCurrentTime()) {
			return false;
		}
		try {
			_connection.do_timestep();

			try {
				synchronized (_positons) {
					_positons.clear();

					int temp = (Integer) _connection.do_job_get(Simulation.getCurrentTime());

					_step = temp / 1000.0;

					if (_start == -1) {
						_start = _step;
					}

					Map<String, VehicleInformationContainer> vehiclePositions = new HashMap<>();

					List<String> allVehicles = requestAllVehicles();
					for (String vehicle : allVehicles) {
						Location position = requestVehiclePosition(vehicle);
						if (position != null) {
							double heading = requestVehicleHeading(vehicle);
							double speed = requestVehicleSpeed(vehicle);
							RoadNetworkRoute route = requestVehicleRoute(vehicle);

							VehicleInformationContainer informationContainer = new VehicleInformationContainer(position, heading, speed, route);

							vehiclePositions.put(vehicle, informationContainer);
						}
					}
					_positons = vehiclePositions;
				}
			} catch (Exception e) {
				e.printStackTrace();
			}

			return true;
		} catch (RuntimeException e) {
			throw e;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return false;
	}

	public String getVehicleType(String pVehicleID) {
		String vehicleType = "default";
		SumoCommand typeCommand = Vehicle.getTypeID(pVehicleID);
		Object typeObject = requestObject(typeCommand);

		if (typeObject != null) {
			vehicleType = (String) typeObject;
		}

		return  vehicleType;
	}

	@Override
	public void setVehicleSpeed(String pVehicleID, double pVehicleSpeed) {
		SumoCommand typeCommand = Vehicle.setSpeed(pVehicleID, pVehicleSpeed);
		execute(typeCommand);
	}

	@Override
	public void setVehicleGap(String pVehicleID, double gap) {
	    SumoCommand typeCommand = Vehicle.setMinGap(pVehicleID, gap);
	    execute(typeCommand);
	}

    @Override
    public RoadNetworkLane getVehicleLaneId(String pVehicleID) {
	    RoadNetworkLane roadNetworkLane = null;
	    String laneId = "";
        SumoCommand laneCommand = Vehicle.getLaneID(pVehicleID);
        Object laneObject = requestObject(laneCommand);

        if (laneObject != null) {
            laneId = (String) laneObject;
            roadNetworkLane = new RoadNetworkLane(laneId);
        }

        return roadNetworkLane;

    }

	@Override
	public List<String> getNextTLS(String vehicleId) {
		SumoCommand nextTLSCommand = Vehicle.getNextTLS(vehicleId);
		Object nextTLSObject = requestObject(nextTLSCommand);
		SumoNextTLS sumoNextTLS = (SumoNextTLS) nextTLSObject;

		List<String> nextTLSList = new LinkedList<>();
		for (SumoNextTLS.NextTLS nextTLS : sumoNextTLS.ll) {
			nextTLSList.add(nextTLS.tlsID);
		}

		return nextTLSList;
	}

	/*
	@Override
	public long getWaitingTime(String vehicleID) {
		SumoCommand waitingTimeCommand = Vehicle.getWaitingTime(vehicleID);
		Object waitingTimeObject = requestObject(waitingTimeCommand);
		String waitingTime = (String)waitingTimeObject;

		return Long.parseLong(waitingTime);
	}
	*/

	@Override
	public void setVehicleMaxSpeed(String pVehicleID, double pVehicleSpeed) {
		SumoCommand maxSpeedCommand = Vehicle.setMaxSpeed(pVehicleID, pVehicleSpeed);
		execute(maxSpeedCommand);
	}

	private Location requestVehiclePosition(String pVehicleID) {
		if (_vehiclesOutOfRange.containsKey(pVehicleID)) {
			if (_vehiclesOutOfRange.get(pVehicleID) < _step) {
				return null;
			} else {
				_vehiclesOutOfRange.remove(pVehicleID);
			}
		}

		SumoCommand positionCommand = Vehicle.getPosition(pVehicleID);

		Object positionObject = requestObject(positionCommand);

		SumoPosition2D sumoPosition = (SumoPosition2D) positionObject;

		if (_observedAreaSet) {
			if (_startX <= sumoPosition.x && sumoPosition.x <= _endX && _startY <= sumoPosition.y && sumoPosition.y <= _endY) {
				return new PositionVector(sumoPosition.x - _startX, sumoPosition.y - _startY, 0);
			} else {
				double diffX = _startX - sumoPosition.x;
				if (diffX < 0 || sumoPosition.x - _endX < diffX) {
					diffX = sumoPosition.x - _endX;
				}
				if (diffX < 0) {
					diffX = 0;
				}
				double diffY = _startY - sumoPosition.y;
				if (diffY < 0 || sumoPosition.y - _endY < diffY) {
					diffY = sumoPosition.y - _endY;
				}
				if (diffY < 0) {
					diffY = 0;
				}
				double diff = Math.sqrt(Math.pow(diffX, 2) + Math.pow(diffY, 2));

				double timeTillNextJoin = diff / 50;

				//_vehiclesOutOfRange.put(pVehicleID, _step + timeTillNextJoin);

				return null;
			}
		} else {
			return new PositionVector(sumoPosition.x, sumoPosition.y, 0);
		}
	}

	private double requestVehicleHeading(String pVehicleID) {
		if (_vehiclesOutOfRange.containsKey(pVehicleID)) {
			if (_vehiclesOutOfRange.get(pVehicleID) < _step) {
				return -1;
			} else {
				_vehiclesOutOfRange.remove(pVehicleID);
			}
		}

		SumoCommand angleCommand = Vehicle.getAngle(pVehicleID);

		Object angleObject = requestObject(angleCommand);


		if (angleObject != null) {
			return (Double) angleObject;
		}
		return -1;
	}

	private double requestVehicleSpeed(String pVehicleID) {
		if (_vehiclesOutOfRange.containsKey(pVehicleID)) {
			if (_vehiclesOutOfRange.get(pVehicleID) < _step) {
				return -1;
			} else {
				_vehiclesOutOfRange.remove(pVehicleID);
			}
		}

		SumoCommand speedCommand = Vehicle.getSpeed(pVehicleID);

		Object speedObject = requestObject(speedCommand);


		if (speedObject != null) {
			return (Double) speedObject;
		}
		return -1;
	}

	private List<String> requestAllVehicles() {
		SumoCommand idList = Vehicle.getIDList();
		Object object = requestObject(idList);
		SumoStringList list = (SumoStringList)object;

		List<String> result = new ArrayList<>();

		for (String vehicle : list) {
			result.add(vehicle);
		}

		return result;
	}

	@Override
	public List<String> getAllVehicles() {
		return getAllVehicles(_step);
	}

	@Override
	public List<String> getAllVehicles(double pStep) {
		if (pStep == _step) {
			return new ArrayList<>(_positons.keySet());
		}
		throw new UnsupportedOperationException("Future locations is not supported anymore!");
	}

	public RoadNetworkEdge getCurrentEdge(String pVehicleID) {
		obtainRoadNetwork();

		SumoCommand roadIDCommand = Vehicle.getRoadID(pVehicleID);
		String currentRoadID = (String) requestObject(roadIDCommand);

		RoadNetworkEdge edge = _roadNetwork.getEdge(currentRoadID);
		while (edge.isInternal()) {
			if (edge.getAccessibleEdges().size() == 1) {
				edge = edge.getAccessibleEdges().get(0);
			} else {
				break;
			}
		}

		return edge;
	}

	public RoadNetworkRoute requestVehicleRoute(String pVehicleID) {
		obtainRoadNetwork();

		synchronized (_positons) {
			List<RoadNetworkEdge> streets = new ArrayList<>();

			SumoCommand routeCommand = Vehicle.getRoute(pVehicleID);
			Object object = requestObject(routeCommand);
			SumoStringList streetList = (SumoStringList) object;

			RoadNetworkEdge currentEdge = getCurrentEdge(pVehicleID);

			if (currentEdge == null) {
				return null;
			}

			boolean add = false;
			for (String street : streetList) {
				if (street.equals(currentEdge.getEdgeID())) {
					add = true;
				}
				if (add) {
					streets.add(_roadNetwork.getEdge(street));
				}
			}


			if (streets.size() == 0) {
				return new RoadNetworkRoute(new ArrayList<>());
			}

			return new RoadNetworkRoute(streets);
		}
	}

	@Override
	public RoadNetworkRoute getCurrentRoute(String pVehicleID) {
		if (_positons.containsKey(pVehicleID)) {
			VehicleInformationContainer route = _positons.get(pVehicleID);
			if (route != null) {
				return route.getRoute();
			}
		}
		return null;
	}

	public Object requestObject(SumoCommand routeCommand) {
		Object object = null;
		try {
			object = _connection.do_job_get(routeCommand);
		} catch (RuntimeException e) {
			throw e;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return object;
	}

	public void execute(SumoCommand routeCommand) {
		try {
			_connection.do_job_set(routeCommand);
		} catch (RuntimeException e) {
			throw e;
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public double getStep() {
		return _step;
	}

	@Override
	public double getStart() {
		return _start;
	}

	@Override
	public double getMaximumAvailablePrediction() {
		return getStep();
	}

	@Override
	public void setObservedArea(double pStartX, double pStartY, double pEndX, double pEndY) {
		_startX = pStartX;
		_startY = pStartY;
		_endX = pEndX;
		_endY = pEndY;

		_observedAreaSet = true;
	}

	@Override
	public double getScenarioWidth() {
		SumoCommand netBoundaryCommand = Simulation.getNetBoundary();

		try {
			SumoGeometry netBoundary = (SumoGeometry) _connection.do_job_get(netBoundaryCommand);
			return Math.max(netBoundary.coords.get(1).x - netBoundary.coords.get(0).x, 10);
		} catch (Exception e) {
			//Nothing to do
		}
		return -1;
	}

	@Override
	public double getScenarioHeight() {
		SumoCommand netBoundaryCommand = Simulation.getNetBoundary();

		try {
			SumoGeometry netBoundary = (SumoGeometry) _connection.do_job_get(netBoundaryCommand);
			return Math.max(netBoundary.coords.get(1).y - netBoundary.coords.get(0).y, 10);
		} catch (Exception e) {
			//Nothing to do
		}
		return -1;
	}

	public boolean providesRoadInformation() {
		return true;
	}

	public RoadNetworkEdge getVehicleDestination(String pVehicleID) {
		RoadNetworkRoute roadNetworkRoute = getCurrentRoute(pVehicleID);

		return roadNetworkRoute.getDestination();
	}

	@Override
	public RoadNetwork getRoadNetwork() {
		obtainRoadNetwork();

		return _roadNetwork;
	}

	@Override
	public RoadNetworkRoute findNewRoute(String pVehicle) throws NoAdditionalRouteAvailableException {
		List<RoadNetworkRoute> routes = new ArrayList<>();
		RoadNetworkRoute route = getCurrentRoute(pVehicle);

		if (route.getRoute().isEmpty()) {
			return null;
		}

		routes.add(route);
		RoadNetworkRoute findRoute = _algorithm.findRoute(_roadNetwork, route.getStart(), route.getDestination(), routes);
		if (findRoute == null) {
			throw new NoAdditionalRouteAvailableException(route.getStart(), route.getDestination(), routes);
		}
		return findRoute;
	}

	@Override
	public RoadNetworkRoute findNewRoute(String pVehicle, List<RoadNetworkEdge> pEdgesToAvoid, boolean pKeepDestination) throws NoAdditionalRouteAvailableException, NoExitAvailableException {
		if (pKeepDestination) {
			List<RoadNetworkRoute> routes = new ArrayList<>();
			RoadNetworkRoute route = getCurrentRoute(pVehicle);

			if (route.getRoute().isEmpty()) {
				return null;
			}

			routes.add(route);

			RoadNetworkRoute findRoute = _algorithm.findRoute(_roadNetwork, route.getStart(), route.getDestination(), routes, pEdgesToAvoid);
			if (findRoute == null) {
				throw new NoAdditionalRouteAvailableException(route.getStart(), route.getDestination(), routes);
			}
			return findRoute;
		} else {
			RoadNetworkRoute route = getCurrentRoute(pVehicle);

			if (route.getRoute().isEmpty()) {
				return null;
			}

			List<RoadNetworkEdge> routeList = route.getRoute();

			RoadNetworkEdge lastEdge = null;
			boolean search = false;
			outer:for (int i = routeList.size() - 1; i >= 0; i--) {
				if (search) {
					List<RoadNetworkEdge> accessibleEdges = routeList.get(i).getAccessibleEdges();
					if (accessibleEdges.size() > 1) {
						for (int j = 0; j < accessibleEdges.size(); j++) {
							if (!accessibleEdges.get(j).equals(lastEdge)) {
								List<RoadNetworkEdge> edges = new ArrayList<>();

								for (int k = 0; k <= i; k++) {
									edges.add(routeList.get(k));
								}
								edges.add(accessibleEdges.get(j));

								List<RoadNetworkEdge> newPaths;
								while ((newPaths = edges.get(edges.size() - 1).getAccessibleEdges()).size() > 0) {
									edges.add(newPaths.get(_random.nextInt(newPaths.size())));
								}

								for (RoadNetworkEdge roadNetworkEdge : edges) {
									if (pEdgesToAvoid.contains(roadNetworkEdge)) {
										continue outer;
									}
								}

								return new RoadNetworkRoute(edges);
							}
						}
					}
				}
				search = true;
				lastEdge = routeList.get(i);
			}
			throw new NoExitAvailableException(route.getStart(), route.getDestination());
		}
	}

	@Override
	public void rerouteVehicle(String pVehicle, RoadNetworkRoute pRoute) {
		SumoStringList routeEdges = new SumoStringList();
		for (RoadNetworkEdge edge : pRoute.getRoute()) {
			routeEdges.add(edge.getEdgeID());
		}
		execute(Vehicle.setRoute(pVehicle, routeEdges));
	}

	@Override
	public void stopVehicle(String pVehicle) {
		SumoCommand stopCommand = Vehicle.setSpeed(pVehicle, 0);
		execute(stopCommand);
	}

	public void obtainRoadNetwork() {
		if (_roadNetwork == null) {
			if (TEMP_FILE.exists()) {
				try {
					ObjectInputStream inputStream = new ObjectInputStream(new FileInputStream(TEMP_FILE));

					Object readObject = inputStream.readObject();

					if (readObject instanceof SerializableRoadNetwork) {
						SerializableRoadNetwork serializedRoadNetwork = (SerializableRoadNetwork) readObject;
						_roadNetwork = new RoadNetwork(serializedRoadNetwork, this);
					}

					inputStream.close();
				} catch (IOException | ClassNotFoundException | NullPointerException e) {
					//Nothing to do
				}
			}

			if (_roadNetwork != null) {
				System.out.println("Got network from cache");

				SumoCommand edgeIDCommand = Edge.getIDList();

				SumoStringList edgeIDStringList = (SumoStringList) requestObject(edgeIDCommand);

				if (_roadNetwork.getAvailableEdges().size() == edgeIDStringList.size()) {
					boolean matching = true;

					for (String edgeID : edgeIDStringList) {
						if (_roadNetwork.getEdge(edgeID) == null) {
							matching = false;
						}
					}

					if (matching) {
						return;
					}
				}
			}

			SumoCommand laneIDCommand = Lane.getIDList();

			Map<String, RoadNetworkEdge> roadNetwork = new HashMap<>();

			SumoStringList laneIDStringList = (SumoStringList) requestObject(laneIDCommand);

			//Requesting all lanes from sumo.
			for (String laneID : laneIDStringList) {
				SumoCommand edgeIDCommand = Lane.getEdgeID(laneID);

				SumoLinkList linkStringList = (SumoLinkList) requestObject(Lane.getLinks(laneID));

				double angle = getLaneAngle(laneID);

				double maxSpeed = getMaxSpeed(laneID);

				if (linkStringList.size() > 0) {
					String edgeID = (String) requestObject(edgeIDCommand);

					if (!roadNetwork.containsKey(edgeID)) {
						roadNetwork.put(edgeID, new RoadNetworkEdge(edgeID, angle, this));
					}

					RoadNetworkEdge edge = roadNetwork.get(edgeID);

					edge.setOriginalMaxSpeed(maxSpeed);
					edge.addLane(new RoadNetworkLane(laneID));

					for (SumoLink link : linkStringList) {
						String notInternalLane = link.notInternalLane;
						String connectedEdge = (String) requestObject(Lane.getEdgeID(notInternalLane));

						double linkAngle = getLaneAngle(laneID);

						if (!roadNetwork.containsKey(connectedEdge)) {
							RoadNetworkEdge roadNetworkEdge = new RoadNetworkEdge(connectedEdge, linkAngle, this);
							roadNetwork.put(connectedEdge, roadNetworkEdge);
						}

						edge.addConnectedEdge(roadNetwork.get(connectedEdge));
					}
				}


			}
			_roadNetwork = new RoadNetwork(roadNetwork, this, true);

			try {
				ObjectOutputStream outputStream = new ObjectOutputStream(new FileOutputStream(TEMP_FILE));

				outputStream.writeObject(new SerializableRoadNetwork(_roadNetwork));

				outputStream.flush();
				outputStream.close();
			} catch (IOException e) {
				//Nothing to do
				e.printStackTrace();
			}
		}

		RoadNetwork.CURRENT_ROAD_NETWORK = _roadNetwork;
	}

	public double getMaxSpeed(String laneID) {
		SumoCommand maxSpeedCommand = Lane.getMaxSpeed(laneID);
		double maxSpeed = (Double) requestObject(maxSpeedCommand);
		return maxSpeed;
	}

	public double getLaneAngle(String laneID) {
		SumoCommand shapeCommand = Lane.getShape(laneID);
		SumoGeometry geometry = (SumoGeometry) requestObject(shapeCommand);

		LinkedList<SumoPosition2D> coords = geometry.coords;
		SumoPosition2D start = coords.getFirst();
		SumoPosition2D end = coords.getLast();

		double sin = (end.y - start.y) / Math.sqrt(Math.pow(end.x - start.x, 2) + Math.pow(end.y - start.y, 2));

		double angle = Math.asin(sin) * 180 / Math.PI;
		if (end.x - start.x < 0) {
			angle += 180;
		}
		return angle;
	}

	@Override
	public List<Location> getLaneShape(String pLaneID) {
		List<Location> positions = new ArrayList<>();

		boolean set = true;
		
		SumoCommand laneShapeCommand = Lane.getShape(pLaneID);
		SumoGeometry geometry = (SumoGeometry)requestObject(laneShapeCommand);
		for (SumoPosition2D location : geometry.coords) {
			if (!isObservedAreaSet()) {
				positions.add(new PositionVector(location.x, location.y));
			} else {
				if (_startX <= location.x && location.x <= _endX && _startY <= location.y && location.y <= _endY) {
					positions.add(new PositionVector(location.x - _startX, location.y - _startY));
				} else {
					set = true;
				}
			}
		}

		if (set) {
			return positions;
		} else {
			return Collections.emptyList();
		}
	}

	@Override
	public Location getEdgeMeanPosition(String pEdgeID) {
		List<Location> positions = new ArrayList<>();

		for (RoadNetworkLane lane : _roadNetwork.getEdge(pEdgeID).getLanes()) {
			String laneID = lane.getLaneID();
			positions.addAll(getLaneShape(laneID));
		}

		double x = 0;
		double y = 0;
		int count = 0;

		for (Location position : positions) {
			x += position.getLongitude();
			y += position.getLatitude();
			count++;
		}

		return new PositionVector(x / (count), y / (count));
	}

	@Override
	public double getVehicleSpeed(String pVehicleID) {
		SumoCommand speedCommand = Vehicle.getSpeed(pVehicleID);
		Object object = requestObject(speedCommand);

		return (Double)object;
	}

	@Override
	public double getLastStepMeanSpeed(String pEdgeID) {
		SumoCommand speedCommand = Edge.getLastStepMeanSpeed(pEdgeID);
		Object object = requestObject(speedCommand);

		return (Double)object;
	}

	@Override
	public void setEdgeMaxSpeed(String pEdgeID, double pMaxSpeed) {
		SumoCommand setMaxSpeedCommand = Edge.setMaxSpeed(pEdgeID, pMaxSpeed);

		execute(setMaxSpeedCommand);

		RoadNetworkEdge edge = getRoadNetwork().getEdge(pEdgeID);

		edge.setMaxSpeed(pMaxSpeed);
	}

	@Override
	public double getEdgeLength(String pEdgeID) {
		double length = 0;
		for (RoadNetworkLane lane : _roadNetwork.getEdge(pEdgeID).getLanes()) {
			SumoCommand speedCommand = Lane.getLength(lane.getLaneID());
			Object object = requestObject(speedCommand);
			length += (double) object;

		}

		return length / (_roadNetwork.getEdge(pEdgeID).getLaneAmount());
	}

	@Override
	public boolean isEdgeUsable(String pEdgeID) {
		if (_observedAreaSet) {
			List<RoadNetworkLane> lanes = _roadNetwork.getEdge(pEdgeID).getLanes();
			if (lanes.size() > 0) {
				List<Location> laneShape = getLaneShape(lanes.get(0).getLaneID());
				if (laneShape.size() > 1) {
					return true;
				}
			}
			return false;
		}
		return true;
	}

	/**
	 * @return the startX
	 */
	@Override
	public double getStartX() {
		return _startX;
	}

	/**
	 * @return the startY
	 */
	@Override
	public double getStartY() {
		return _startY;
	}

	/**
	 * @return the endX
	 */
	@Override
	public double getEndX() {
		return _endX;
	}

	/**
	 * @return the endY
	 */
	@Override
	public double getEndY() {
		return _endY;
	}

	/**
	 * @return the observedAreaSet
	 */
	@Override
	public boolean isObservedAreaSet() {
		return _observedAreaSet;
	}

    @Override
    public List<TrafficLight> getAllTrafficLights() {
	    SumoCommand trafficLightCommand = Trafficlight.getIDList();
	    Object object = requestObject(trafficLightCommand);
	    SumoStringList trafficLightList = (SumoStringList)object;

	    List<TrafficLight> result = new ArrayList<>();

	    for (String trafficLightId : trafficLightList) {
	        TrafficLight trafficLight = new TrafficLight(trafficLightId, this);
	        List<String> controlledLanesIds = trafficLight.getControlledLaneIDs();

	        boolean found = false;
	        for (String laneId : controlledLanesIds) {
	        	if (isLaneUsable(laneId)) {
	        		found = true;
	        		break;
				}
			}

			if (found) {
	        	result.add(trafficLight);
			}
		}

		return result;
    }

    @Override
    public List<String> getControlledLanes(String pTrafficLightID) {
        SumoCommand trafficLightCommand = Trafficlight.getControlledLanes(pTrafficLightID);
        Object object = requestObject(trafficLightCommand);
        SumoStringList list = (SumoStringList)object;

		List<String> result = new ArrayList<>();

        for (String lane : list) {
        	result.add(lane);
		}

		return result;
    }

	@Override
	public List<String> getControlledEdges(String pTrafficLightID) {
		List<String> controlledLanesList = getControlledLanes(pTrafficLightID);
		Map<String, String> controlledEdgeHash = new HashMap<>();
		List<String> controlledEdgeList = new ArrayList<>();

		controlledLanesList.removeIf(lane -> lane.charAt(0) == ':');

		for (String lane : controlledLanesList) {
			String edge = lane.split("_")[0];
			if (!controlledEdgeHash.containsKey(edge)) {
				controlledEdgeHash.put(edge, "");
			}
		}

		for (String edge : controlledEdgeHash.keySet()) {
			controlledEdgeList.add(edge);
		}

		return controlledEdgeList;
	}

	@Override
    public Map<RoadNetworkLink, TrafficLightState> getTrafficLightStateForEachLane(String pTrafficLightID) {
	    SumoCommand trafficLightCommand = Trafficlight.getRedYellowGreenState(pTrafficLightID);
	    Object trafficLightObject = requestObject(trafficLightCommand);
	    String trafficLightStates = (String)trafficLightObject;

	    Map<RoadNetworkLink, TrafficLightState> result = new HashMap<>();
	    List<RoadNetworkLink> controlledLinks = getControlledLinks(pTrafficLightID);

	    for (int i = 0 ; i < controlledLinks.size() ; ++i) {
	    	switch (trafficLightStates.toLowerCase().charAt(i)) {
				case 'g':
					result.put(controlledLinks.get(i), TrafficLightState.GREEN);
					break;
				case 'y':
					result.put(controlledLinks.get(i), TrafficLightState.YELLOW);
					break;
				case 'r':
					result.put(controlledLinks.get(i), TrafficLightState.RED);
					break;
				case 'o':
					result.put(controlledLinks.get(i), TrafficLightState.OFF);
					break;
				default:
					throw new AssertionError("Unexpected value!");
			}
		}

        return result;
    }

	public List<RoadNetworkLink> getControlledLinks(String pTrafficLightID) {
		SumoCommand trafficLightCommand = Trafficlight.getControlledLinks(pTrafficLightID);
		Object controlledLinksObject = requestObject(trafficLightCommand);
		SumoLinkList controlledLanes = (SumoLinkList)controlledLinksObject;

		List<RoadNetworkLink> result = new ArrayList<>();

		for (SumoLink link : controlledLanes) {
			RoadNetworkLink roadNetworkLink = new RoadNetworkLink(link.from, link.to);
			result.add(roadNetworkLink);
		}

		return result;
	}

	private boolean isLaneUsable(String laneID) {
		List<Location> laneShape = getLaneShape(laneID);
		for (Location location : laneShape) {
			if (0 <= location.getLongitude() && location.getLongitude() <= _endX - _startX && 0 <= location.getLatitude() && location.getLatitude() <= _endY - _startY) {
				return true;
			}
		}
		return false;
	}

	@Override
	public int getMinExpectedNumber() {
		SumoCommand simulationCommand = Simulation.getMinExpectedNumber();
		Object objectSimulation = requestObject(simulationCommand);
		int expectedNumber = (int) objectSimulation;

		return expectedNumber;
	}
}
